#!/usr/bin/env python
#coding=utf-8


"""
Port forwarder with graceful exit.

Run the example as

  python portforwarder.py :8080 gevent.org:80

Then direct your browser to http://localhost:8080 or do "telnet localhost 8080".

When the portforwarder receives TERM or INT signal (type Ctrl-C),
it closes the listening socket and waits for all existing
connections to finish. The existing connections will remain unaffected.
The program will exit once the last connection has been closed.

"""


from __future__ import print_function, unicode_literals, absolute_import

import os
import os.path
import sys
import signal
import argparse
import logging
import gevent
from gevent.server import StreamServer
from gevent.socket import create_connection, gethostbyname


__version__ = "0.2.1"


#----------------------------------------------------------------------
def forward(source, dest):
    """forward loop"""
    try:
        while True:
            data = source.recv(1024)
            if not data:
                break
            dest.sendall(data)
    except:
        source.close()
        dest.close()

    return


########################################################################
class PortForwarder(object):

    forwarde_table = {}

    #----------------------------------------------------------------------
    def __init__(self, source=None, dest=None, conf_filename=None):
        """Constructor"""
        if conf_filename is not None:
            #config file mode
            self.parse_config_file(conf_filename)

        elif source and dest:
            #direct mode
            self.add_forward_pair(source, dest)

        else:
            logging.error('init portforwarder ERROR!')

        return

    #----------------------------------------------------------------------
    def parse_config_file(self, filename):
        """parse config file"""
        try:
            with open(filename, mode='r') as config_f:
                for line in iter(config_f.readline, ''):
                    if line == '' or line[0] == '#':
                        continue

                    source, dest = line.split(' ')
                    self.add_forward_pair(source, dest)
        except IOError:
            print('can not open/parse config file: [%s]' % filename)
            print('program exit!')
            exit(1)

        return

    #----------------------------------------------------------------------
    def add_forward_pair(self, source, dest):
        """add and bind one forward pair"""
        source_address = self.parse_address(source)
        dest_address = self.parse_address(dest)

        if source_address in self.forwarde_table:
            return False
        else:
            logging.info('Add forward pair %s to %s', str(source_address), str(dest_address))
            self.forwarde_table[source_address] = {
                'source_address': source_address,
                'dest_address': dest_address,
                'server': StreamServer(source_address, self.handle),
            }

        return True

    #----------------------------------------------------------------------
    def start(self):
        """start server"""
        if len(self.forwarde_table) == 0:
            logging.error('no forward setting!')
            return

        gevent.signal(signal.SIGTERM, self.close)
        gevent.signal(signal.SIGINT, self.close)

        for k in self.forwarde_table:
            self.forwarde_table[k]['server'].start()

        gevent.wait()
        return

    #----------------------------------------------------------------------
    def parse_address(self, address):
        """
        parse address string to tuple.

        from: '127.0.0.1:8080'
        to: ('127.0.0.1', 8080)
        """
        try:
            hostname, port = address.rsplit(':', 1)
            port = int(port)

            if hostname == '' or hostname == 'localhost':
                hostname = '127.0.0.1'

        except ValueError:
            sys.exit('Expected HOST:PORT: %r' % address)
        return gethostbyname(hostname), port

    #----------------------------------------------------------------------
    def handle(self, source_socket, address):
        """listener handle func"""
        source_address = source_socket.getsockname()
        logging.info('forward request from:%s to:%s accepted', str(address[:2]), str(source_address))

        try:
            dest_address = self.forwarde_table[source_address]['dest_address']

            dest_socket = create_connection(dest_address)
        except IOError as ex:
            logging.warning('%s:%s failed to connect to %s:%s: %s', address[0], address[1], dest_address[0], dest_address[1], ex)
            return
        gevent.spawn(forward, source_socket, dest_socket)
        gevent.spawn(forward, dest_socket, source_socket)

        return

    #----------------------------------------------------------------------
    def close(self):
        """close listener"""
        logging.info('port forwarder closing...')

        closed = False
        for k in self.forwarde_table:
            if self.forwarde_table[k]['server'].closed:
                closed = True
            else:
                self.forwarde_table[k]['server'].close()
                self.forwarde_table[k]['server'].stop()

        if closed:
            logging.info('port forwarder force exit!')
            #force break forward()
            exit()

        return


#----------------------------------------------------------------------
def try_fork():
    """"""
    try:
        pid = os.fork()
        if pid > 0:
            # Exit first parent
            sys.exit(0)
    except OSError as err:
        print('fork error!')
        print(err)
        sys.exit(1)

    return


#----------------------------------------------------------------------
def cmd():
    """CLI args parse"""
    parser = argparse.ArgumentParser()
    parser.add_argument('--address', '-a', nargs=2, metavar=('source-address', 'dest-address'), help="e.g. python portforwarder.py :8080 localhost:80")

    parser.add_argument('--filename', '-f', type=str, help="config file name")

    parser.add_argument('--daemon', '-d', action='store_true', help="trun on daemon mode")

    return parser.parse_args()


#----------------------------------------------------------------------
def main():
    """main func"""
    args = cmd()

    if args.daemon:
        try_fork()

    logger_level = logging.INFO
    logging.basicConfig(filename='portforwarder.log', format='%(asctime)s %(levelname)s:%(message)s', level=logger_level)

    if args.address == None and args.filename == None:
        args.filename = os.path.join(os.path.expanduser('~'), '.config', 'portforworder', 'portforworder.conf')
        pf = PortForwarder(None, None, args.filename)

    elif args.address == None:
        pf = PortForwarder(None, None, args.filename)

    else:
        pf = PortForwarder(args.address[0], args.address[1], None)

    pf.start()
    return


if __name__ == '__main__':
    main()

